/*************************************************************************

      Graphics Research Group - Department of Computer Science
                  University of Sheffield, UK
               Michael Meredith & Steve Maddock
                     Copyright (c) 2010
                    All Rights Reserved.


  Permission is hereby granted, free of charge, to use this software 
  and its documentation without restriction, including without limitation 
  the rights to use, copy, modify, merge and publish, subject to the following 
  conditions:

   1. The code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Any modifications must be clearly marked as such.

   3. Original authors' names are not deleted.

   4. The authors' names are not used to endorse or promote products
      derived from this software without specific prior written
      permission from the author.

   5. This software is used in a research, non-profit capacity only and not 
      for commercial or profit applications unless specific prior written
      permission from the author is given.


  THE UNIVERSITY OF SHEFFIELD AND THE CONTRIBUTORS TO THIS WORK
  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
  SHALL THE UNIVERSITY OF SHEFFIELD NOR THE CONTRIBUTORS BE LIABLE
  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
  THIS SOFTWARE.

*************************************************************************/

import java.io.*;
import java.util.StringTokenizer;

public class BVHReader
{
  private static final int UNKNOWN=0;
  private static final int HIERARCHY=1;
  private static final int MOTION=2;
  
  public static MocapData loadBVHFile(String filename)
  {
    int lineno=0;
    int section=UNKNOWN;
    boolean endsite=false;
    
    MocapData mocap=new MocapData();
    Node curnode=null;
    float fdata[]=new float[6];
    int frameno=0;
    
    try
    {
      BufferedReader file=new BufferedReader(new FileReader(filename));
      String data=file.readLine();
      
      while (data!=null)
      {
        ++lineno;
        if (section==HIERARCHY)
        {
          StringTokenizer tokens=new StringTokenizer(data, " \t");
          String val=tokens.nextToken();
          
          if (val.equals("{")) // this clause can be ignored
          {
          }
          else if (val.equals("}"))
          {
            if (endsite)
              endsite=false;
            else  if (curnode!=null)
              curnode=curnode.getParent();
            else
              BVHFileError("going back a level without a parent", lineno);
          }
          else if (tokens.hasMoreTokens())
          {
            if (val.equalsIgnoreCase("ROOT"))
            {
              if (mocap.getRoot()==null)
              {
                mocap.addNode(curnode=new Node(tokens.nextToken()));
                mocap.setRoot(curnode);
              }
              else
                BVHFileError("only one root entry is allowed", lineno);
            }
            else if (val.equalsIgnoreCase("JOINT"))
            {
              if (curnode!=null)
                mocap.addNode(curnode=curnode.addChild(tokens.nextToken()));
              else
                BVHFileError("no parent node for JOINT", lineno);
            }
            else if (val.equalsIgnoreCase("OFFSET"))
            {
              if (curnode!=null)
              {
                float x, y, z;
                boolean error=true;

                x=Float.valueOf(tokens.nextToken()).floatValue();
                y=z=0;
                if (tokens.hasMoreTokens())
                {
                  y=Float.valueOf(tokens.nextToken()).floatValue();
                  if (tokens.hasMoreTokens())
                  {
                    z=Float.valueOf(tokens.nextToken()).floatValue();
                    error=false;
                  }
                }
                
                if (error)
                  BVHFileError("not enough offset data", lineno);
                else
                {
                  if (endsite)
                    curnode.setLength(x, y, z);
                  else
                  {
                    curnode.setOffset(x, y, z);
                    
                    // the offset of the child node is used to determine the length of the parent node if it is the first child
                    Node parent=curnode.getParent();
                    if (parent!=null && parent.getNoOfChildren()==1)
                      parent.setLength(x, y, z);
                  }
                }
              }
              else
                BVHFileError("no node for offset data", lineno);
            }
            else if (val.equalsIgnoreCase("CHANNELS"))
            {
              if (curnode!=null)
              {
                boolean error=false;
                int channels=Integer.valueOf(tokens.nextToken()).intValue();
                int[] doforder=new int[channels];

                for (int i=0; i<channels && !error; i++)
                {
                  if (tokens.hasMoreTokens())
                  {
                    val=tokens.nextToken();
                    if (val.equalsIgnoreCase("Xposition"))
                      doforder[i]=Node.XTRANSLATION;
                    else if (val.equalsIgnoreCase("Yposition"))
                      doforder[i]=Node.YTRANSLATION;
                    else if (val.equalsIgnoreCase("Zposition"))
                      doforder[i]=Node.ZTRANSLATION;
                    else if (val.equalsIgnoreCase("Xrotation"))
                      doforder[i]=Node.XROTATION;
                    else if (val.equalsIgnoreCase("Yrotation"))
                      doforder[i]=Node.YROTATION;
                    else if (val.equalsIgnoreCase("Zrotation"))
                      doforder[i]=Node.ZROTATION;
                  }
                  else
                    error=true;
                }
                
                if (error)
                  BVHFileError("not enough channel data", lineno);
                else
                  curnode.setDOFOrder(doforder);
              }
              else
                BVHFileError("no node for channel data", lineno);
            }
            else if (val.equalsIgnoreCase("END") && tokens.nextToken().equalsIgnoreCase("SITE"))
              endsite=true;
          }
        }
        else if (section==MOTION)
        {
          StringTokenizer tokens=new StringTokenizer(data, " \t");
          String val=tokens.nextToken();

          if (tokens.hasMoreTokens())
          {
            if (val.equalsIgnoreCase("Frames:"))
            {
              mocap.setNoOfFrames(frameno=Integer.valueOf(tokens.nextToken()).intValue());
              
              for (int i=mocap.getNoOfNodes()-1; i>=0; --i)
                mocap.getNode(i).setNoOfFrames(frameno);

              frameno=0;    // reset frameno count, which is used when reading in the motion data
            }
            else if (val.equalsIgnoreCase("Frame") && tokens.nextToken().equalsIgnoreCase("Time:") && tokens.hasMoreTokens())
              mocap.setFrameRate(1.0f/Float.valueOf(tokens.nextToken()).floatValue());
            else
            {
              boolean error=false;
              
              // Restart tokenizer because we have already taken the first element
              tokens=new StringTokenizer(data, " \t");
              int len=mocap.getNoOfNodes();
              for (int i=0; i<len && !error; i++)
              {
                curnode=mocap.getNode(i);
                int noofdofs=curnode.getNoOfDOFs();
                for (int j=0; j<noofdofs && !error; j++)
                {
                  if (tokens.hasMoreTokens())
                    fdata[j]=Float.valueOf(tokens.nextToken()).floatValue();
                  else
                    error=true;
                }
                
                curnode.setFrameData(frameno, fdata);
              }
              
              if (error)
                BVHFileError("not enough frame data", lineno);
              ++frameno;
            }
            
          }
        }
        
        if (data!=null)
        {
          if (data.equalsIgnoreCase("HIERARCHY"))
            section=HIERARCHY;
          else if (data.equalsIgnoreCase("MOTION"))
            section=MOTION;
        }
        data=file.readLine();
      }
    
      return mocap;
    }
    catch (Exception e)
    {
      System.out.println("An error occured while reading in the specified file: "+e);
      return null;
    }
  }
  
  private static void BVHFileError(String error, int lineno)
  {
    System.out.println("error in mocap file - line "+lineno+": "+error);
    System.exit(0);
  }
}