/*************************************************************************

      Graphics Research Group - Department of Computer Science
                  University of Sheffield, UK
               Michael Meredith & Steve Maddock
                     Copyright (c) 2010
                    All Rights Reserved.


  Permission is hereby granted, free of charge, to use this software 
  and its documentation without restriction, including without limitation 
  the rights to use, copy, modify, merge and publish, subject to the following 
  conditions:

   1. The code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Any modifications must be clearly marked as such.

   3. Original authors' names are not deleted.

   4. The authors' names are not used to endorse or promote products
      derived from this software without specific prior written
      permission from the author.

   5. This software is used in a research, non-profit capacity only and not 
      for commercial or profit applications unless specific prior written
      permission from the author is given.


  THE UNIVERSITY OF SHEFFIELD AND THE CONTRIBUTORS TO THIS WORK
  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
  SHALL THE UNIVERSITY OF SHEFFIELD NOR THE CONTRIBUTORS BE LIABLE
  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
  THIS SOFTWARE.

*************************************************************************/

import java.nio.*;
import java.awt.*;
import java.awt.event.*; 

import javax.media.opengl.*;
import javax.media.opengl.glu.*;
import com.sun.opengl.util.GLUT;

public class MocapDisplay extends Frame implements ActionListener 
{
  private GLSpace glspace;
  private GLCanvas glcanvas;
  
  public static void main(String[] args)
  {
    MocapDisplay gl=new MocapDisplay();
    gl.setVisible(true);
  }
    
  public MocapDisplay() 
  {
    super("MocapDisplay");
    setSize(640, 480);
        
    addWindowListener(new WindowAdapter() 
    {
      public void windowClosing(WindowEvent e) 
      {
        System.exit(0);
      }
      public void windowOpened(WindowEvent e) 
      {
        FPSAnimator animator=new FPSAnimator(glcanvas, 60);
        animator.start();
      }
    } );

    MenuBar menuBar=new MenuBar();
    setMenuBar(menuBar);
      Menu fileMenu=new Menu("File");
        MenuItem quitItem=new MenuItem("Quit");
        quitItem.addActionListener(this);
        fileMenu.add(quitItem);
    menuBar.add(fileMenu);

    glspace=new GLSpace();
    glcanvas=new GLCanvas();
    glcanvas.addGLEventListener(glspace);

    add(glcanvas, "Center");

    glcanvas.addMouseMotionListener(glspace);
  }

  public void actionPerformed(ActionEvent e) 
  {
    if (e.getActionCommand().equalsIgnoreCase("quit"))
      System.exit(0);
  }
}

class GLSpace implements GLEventListener, MouseMotionListener
{
  private static final float NEAR_CLIP=0.1f;
  private static final float FAR_CLIP=100.0f;
  
  private Point lastpoint;
  private float[] eye, upvec;
  private float radius, theta, phi;
  private int width, height;

  private GLU glu = new GLU();
  private GLUT glut = new GLUT();

  private MocapData mocap;
  private long starttime;
  
  public GLSpace() 
  {
    eye=new float[3];
    upvec=new float[3];
    radius=4.0f;
    theta=(float) (-Math.PI/2.0f);
    phi=(float) (Math.PI/8.0f);
    calcEyePosition();
    lastpoint=new Point();
  }
  
  public void init(GLAutoDrawable drawable) 
  {
    GL gl=drawable.getGL(); 

    gl.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    gl.glEnable(GL.GL_DEPTH_TEST);
    gl.glDisable(GL.GL_CULL_FACE);
    gl.glShadeModel(GL.GL_SMOOTH);
    gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_FILL);
    gl.glEnable(GL.GL_NORMALIZE);
    gl.glEnable(GL.GL_LIGHTING);
    
    createObjects(gl);
    starttime=System.currentTimeMillis();
  }
  
  private void createObjects(GL gl)
  {
    mocap=BVHReader.loadBVHFile("Hop-Mike 1.bvh");
    System.out.println(mocap);
  }

  public void displayChanged(GLAutoDrawable drawable, boolean modeChanged, boolean deviceChanged) {}
  
  public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height)
  {
    GL gl=drawable.getGL();  
    
    this.width=width;
    this.height=height;
    
    float fAspect=(float) width/height;
    float fovy=60.0f;

    gl.glViewport(0, 0, width, height);
    gl.glMatrixMode(GL.GL_PROJECTION);
    gl.glLoadIdentity();
    
    float top=(float) Math.tan(Math.toRadians(fovy*0.5))*NEAR_CLIP;
    float bottom=-top;
    float left=fAspect*bottom;
    float right=fAspect*top;
    
    gl.glFrustum(left, right, bottom, top, NEAR_CLIP, FAR_CLIP);
    gl.glMatrixMode(GL.GL_MODELVIEW);
  }
  
  public void display(GLAutoDrawable drawable) 
  {
    GL gl=drawable.getGL();

    gl.glClear(GL.GL_COLOR_BUFFER_BIT|GL.GL_DEPTH_BUFFER_BIT);
    gl.glLoadIdentity();
    light0(gl);
    glu.gluLookAt(eye[0], eye[1], eye[2], 0.0f, 0.0f, 0.0f, upvec[0], upvec[1], upvec[2]);
    
    displayAxes(gl, glu, 1.0f, 1.0f, 1.0f);
    
    gl.glPushMatrix();
      gl.glScalef(0.03f, 0.03f, 0.03f);
      displayMocap(gl);
    gl.glPopMatrix();
  }
  
  private void displayMocap(GL gl)
  {
    long dur=System.currentTimeMillis()-starttime;
    int frame=(int) ((dur*mocap.getFrameRate())/1000);
    frame%=mocap.getNoOfFrames();
    
    gl.glLineWidth(10.0f);
    gl.glColor3f(1.0f, 1.0f, 1.0f);
    gl.glDisable(GL.GL_LIGHTING);
    displayMocapNode(gl, frame, mocap.getRoot());
    gl.glEnable(GL.GL_LIGHTING);
  }
  
  private void displayMocapNode(GL gl, int frame, Node node)
  {
    gl.glPushMatrix();
      float[] offset1=node.getOffset();
      float[] offset2=node.getFrameOffset(frame);
      gl.glTranslatef(offset1[0]+offset2[0], offset1[1]+offset2[1], offset1[2]+offset2[2]);
      
      float angles[]=node.getFrameAngles(frame);
      int doforder[]=node.getDOFOrder();
      
      // ignore translations because they are assumed to be first and processed above
      for (int i=0; i<doforder.length; i++)
      {
        if (doforder[i]==Node.XROTATION)
          gl.glRotatef(angles[0], 1.0f, 0.0f, 0.0f);
        else if (doforder[i]==Node.YROTATION)
          gl.glRotatef(angles[1], 0.0f, 1.0f, 0.0f);
        else if (doforder[i]==Node.ZROTATION)
          gl.glRotatef(angles[2], 0.0f, 0.0f, 1.0f);
      }
      
      gl.glBegin(GL.GL_LINES);
        gl.glVertex3f(0.0f, 0.0f, 0.0f);
        gl.glVertex3fv(node.getLength(),0);
      gl.glEnd();
      
      gl.glPushMatrix();
        for (int i=0; i<node.getNoOfChildren(); i++)
          displayMocapNode(gl, frame, node.getChild(i));
      gl.glPopMatrix();
      
    gl.glPopMatrix();
  }

  private void light0(GL gl)
  {
    gl.glEnable(GL.GL_LIGHT0);
    float[] lightPosition={0.0f, 4.0f, 0.0f, 0.0f}; 
    float[] ambient={0.0f, 0.0f, 0.0f, 1.0f};
    float[] whiteLight={1.0f, 1.0f, 1.0f, 1.0f};
    gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, lightPosition,0);
    gl.glLightfv(GL.GL_LIGHT0, GL.GL_AMBIENT, ambient,0);
    gl.glLightfv(GL.GL_LIGHT0, GL.GL_DIFFUSE, whiteLight,0);
    gl.glLightfv(GL.GL_LIGHT0, GL.GL_SPECULAR, whiteLight,0);
  }

  public void mouseDragged(MouseEvent e)
  {
    Point ms=e.getPoint();
    
    float dx=(float) (ms.x-lastpoint.x)/width;
    float dy=(float) (ms.y-lastpoint.y)/height;
    
    if (e.getModifiers()==MouseEvent.BUTTON1_MASK)
    {
      theta-=dx*2.0f;
      phi+=dy*2.0f;
    }
    else if (e.getModifiers()==MouseEvent.BUTTON3_MASK)
      radius-=dy;
    
    calcEyePosition();
    
    lastpoint=ms;
  }

  public void mouseMoved(MouseEvent e)
  {   
    lastpoint=e.getPoint(); 
  }
  
  private void calcEyePosition()
  {
    float cy, cz, sy, sz;
    cy=(float) Math.cos(theta);
    sy=(float) Math.sin(theta);
    cz=(float) Math.cos(phi);
    sz=(float) Math.sin(phi);
    
    eye[0]=radius*cy*cz;
    eye[1]=radius*sz;
    eye[2]=-radius*sy*cz;
    
    upvec[0]=-cy*sz;
    upvec[1]=cz;
    upvec[2]=sy*sz;

    // To keep the camera always point upwards, if the y-component of the up vector is negative, inverse the whole up vector    
    if (upvec[1]<0)
    {
      upvec[0]=-upvec[0];
      upvec[1]=-upvec[1];
      upvec[2]=-upvec[2];
    }
  }
  
  private void displayAxes(GL gl, GLU glu, float xLength, float yLength, float zLength)
  {
    // All axes will be set to same specular, shininess and emission settings.
    // Each axis will have will be set to different ambient and diffuse values.
    float[] matXAmbientDiffuse = {0.7f, 0.0f, 0.0f, 1.0f};
    float[] matYAmbientDiffuse = {0.0f, 0.7f, 0.0f, 1.0f};
    float[] matZAmbientDiffuse = {0.0f, 0.0f, 0.7f, 1.0f};
    float[] matSpecular = {1.0f, 1.0f, 1.0f, 1.0f};
    float[] matShininess = {32.0f};
    float[] matEmission = {0.0f, 0.0f, 0.0f, 1.0f};

    float cylinderRadius = 0.05f;
    float coneRadius = 0.05f;
    float coneHeight = 0.2f;
    
    int sphereslices = 20;
    int spherestacks = 20;
    int coneslices = 10;
    int conestacks = 10;
     
    gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE, matXAmbientDiffuse,0);
    gl.glMaterialfv(GL.GL_FRONT, GL.GL_SPECULAR, matSpecular,0);
    gl.glMaterialfv(GL.GL_FRONT, GL.GL_SHININESS, matShininess,0);
    gl.glMaterialfv(GL.GL_FRONT, GL.GL_EMISSION, matEmission,0);

    // x axis is half red
    gl.glPushMatrix();
      gl.glTranslatef(xLength/2.0f, 0.0f, 0.0f);
      gl.glScalef(xLength/2.0f, cylinderRadius, cylinderRadius);
      glut.glutSolidSphere(1.0f, sphereslices, spherestacks);
    gl.glPopMatrix();
    gl.glPushMatrix();
      gl.glTranslatef(xLength, 0.0f, 0.0f);
      gl.glRotatef(90.0f, 0.0f, 1.0f, 0.0f);
      // auxSolidCone(...) is defined with the point along the z axis
      // and the base on the xy plane.
      glut.glutSolidCone(coneRadius, coneHeight, coneslices, conestacks);
    gl.glPopMatrix();

    // y axis is half green
    gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE, matYAmbientDiffuse,0);
    gl.glPushMatrix();
      gl.glTranslatef(0.0f, yLength/2.0f, 0.0f);
      gl.glScalef(cylinderRadius, yLength/2.0f, cylinderRadius);
      glut.glutSolidSphere(1.0f, sphereslices, spherestacks);
    gl.glPopMatrix();
    gl.glPushMatrix();
      gl.glTranslatef(0.0f, yLength, 0.0f);
      gl.glRotatef(-90.0f, 1.0f, 0.0f, 0.0f);
      glut.glutSolidCone(coneRadius, coneHeight, coneslices, conestacks);
    gl.glPopMatrix();

    // z axis is half blue
    gl.glMaterialfv(GL.GL_FRONT, GL.GL_AMBIENT_AND_DIFFUSE, matZAmbientDiffuse,0);
    gl.glPushMatrix();
      gl.glTranslatef(0.0f, 0.0f, zLength/2.0f);
      gl.glScalef(cylinderRadius, cylinderRadius, yLength/2.0f);
      glut.glutSolidSphere(1.0f, sphereslices, spherestacks);
    gl.glPopMatrix();
    gl.glPushMatrix();
      gl.glTranslatef(0.0f, 0.0f, zLength);
      glut.glutSolidCone(coneRadius, coneHeight, coneslices, conestacks);
    gl.glPopMatrix();
  }  
}
