/*************************************************************************

      Graphics Research Group - Department of Computer Science
                  University of Sheffield, UK
               Michael Meredith & Steve Maddock
                     Copyright (c) 2010
                    All Rights Reserved.


  Permission is hereby granted, free of charge, to use this software 
  and its documentation without restriction, including without limitation 
  the rights to use, copy, modify, merge and publish, subject to the following 
  conditions:

   1. The code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Any modifications must be clearly marked as such.

   3. Original authors' names are not deleted.

   4. The authors' names are not used to endorse or promote products
      derived from this software without specific prior written
      permission from the author.

   5. This software is used in a research, non-profit capacity only and not 
      for commercial or profit applications unless specific prior written
      permission from the author is given.


  THE UNIVERSITY OF SHEFFIELD AND THE CONTRIBUTORS TO THIS WORK
  DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
  ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT
  SHALL THE UNIVERSITY OF SHEFFIELD NOR THE CONTRIBUTORS BE LIABLE
  FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
  WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
  AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
  ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF
  THIS SOFTWARE.

*************************************************************************/

public class Node
{
  private String name;
  private float[] length;
  private float[] offset, euler;
  private Node children[], parent;
  private float[][] frameangles, frameoffsets; // frame data will always be stored as [0]=x, [1]=y, [2]=z
  private int doforder[];
  
  public static final int XTRANSLATION=0x1;
  public static final int YTRANSLATION=0x2;
  public static final int ZTRANSLATION=0x4;
  public static final int XROTATION   =0x8;
  public static final int YROTATION   =0x10;
  public static final int ZROTATION   =0x20;
  
  public Node() 
  {
    init();
  }
  
  public Node(String name)
  {
    init();
    this.name=name;
  }
  
  private void init()
  {
    offset=new float[3];
    euler=new float[3];
    length=new float[3];
  }
  
  public String getName()
  {
    return name;
  }
  
  public void setName(String name)
  {
    this.name=name;
  }
  
  public float[] getLength()
  {
    return length;
  }
  
  public void setLength(float x, float y, float z)
  {
    length[0]=x;
    length[1]=y;
    length[2]=z;
  }
  
  public float[] getOffset()
  {
    return offset;
  }
  
  public void setOffset(float x, float y, float z)
  {
    offset[0]=x;
    offset[1]=y;
    offset[2]=z;
  }
  
  public float[] getEuler()
  {
    return euler;
  }
  
  public void setEuler(float x, float y, float z)
  {
    euler[0]=x;
    euler[1]=y;
    euler[2]=z;
  }
  
  public int getNoOfChildren()
  {
    return (children==null?0:children.length);
  }
  
  public Node getChild(int index)
  {
    if (index<getNoOfChildren())
      return children[index];
    else
      return null;
  }
  
  public Node getParent()
  {
    return parent;
  }
  
  public void setParent(Node parent)
  {
    this.parent=parent;
  }
  
  public Node addChild(String name)
  {
    int len=getNoOfChildren();
    Node[] temp=new Node[len+1];
    for (int i=0; i<len; i++)
      temp[i]=children[i];
    children=temp;
    children[len]=new Node(name);
    children[len].setParent(this);
    return children[len];
  }
  
  public float[] getFrameAngles(int frame)
  {
    if (frameangles!=null && frame<frameangles.length)
      return frameangles[frame];
    else
      return null;
  }
  
  public float[] getFrameOffset(int frame)
  {
    if (frameoffsets!=null && frame<frameoffsets.length)
      return frameoffsets[frame];
    else
      return null;
  }

  public int[] getDOFOrder()
  {
    return doforder;
  }
  
  public void setDOFOrder(int[] doforder)
  {
    this.doforder=doforder;
  }
  
  public int getNoOfDOFs()
  {
    return (doforder==null?0:doforder.length);
  }
  
  public void setNoOfFrames(int frames)
  {
    frameangles=new float[frames][3];
    frameoffsets=new float[frames][3];
  }
  
  public void setFrameData(int frameno, float[] fdata)
  {
    if (doforder!=null && frameangles!=null && frameno<frameangles.length)
    {
      int dofs=getNoOfDOFs();
      for (int i=0; i<dofs; i++)
      {
        if (doforder[i]==XTRANSLATION)
          frameoffsets[frameno][0]=fdata[i];
        else if (doforder[i]==YTRANSLATION)
          frameoffsets[frameno][1]=fdata[i];
        else if (doforder[i]==ZTRANSLATION)
          frameoffsets[frameno][2]=fdata[i];
        else if (doforder[i]==XROTATION)
          frameangles[frameno][0]=fdata[i];
        else if (doforder[i]==YROTATION)
          frameangles[frameno][1]=fdata[i];
        else if (doforder[i]==ZROTATION)
          frameangles[frameno][2]=fdata[i];
      }
    }
  }
  
  public String toString()
  {
    return name+"; children="+getNoOfChildren()+"; length="+length[0]+", "+length[1]+", "+length[2];
  }
  
  public String getHierarchy(int level)
  {
    String res="";
    for (int i=0; i<level; i++)
      res+=" ";
    res+=toString()+"\r\n";
    
    for (int i=0; i<getNoOfChildren(); i++)
      res+=children[i].getHierarchy(level+1);
    return res;
  }
}